from __future__ import annotations

import re
from typing import List, Optional

from ._errors import SparkError
from ._utils import is_positive_int, is_str_empty

__all__ = ['Validators']


class BaseValidator:
    _errors: List[SparkError] = []

    @property
    def errors(self):
        return self._errors

    def reset(self):
        self._errors.clear()


class EmptyStringValidator(BaseValidator):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(EmptyStringValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Optional[str], message: Optional[str] = None):
        if is_str_empty(value):
            raise SparkError.sdk(message or 'must be non-empty string value', cause=value)

    def is_valid(self, value: Optional[str], message: Optional[str] = None) -> bool:
        try:
            self.validate(value, message)
            return True
        except SparkError as error:
            self._errors.append(error)
            return False


class PositiveNumberValidator(BaseValidator):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(PositiveNumberValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Optional[int | float]):
        if not is_positive_int(value):
            raise SparkError.sdk('must be a positive number', value)

    def is_valid(self, value: Optional[int | float]) -> bool:
        try:
            self.validate(value)
            return True
        except SparkError as error:
            self._errors.append(error)
            return False


class BaseUrlValidator(BaseValidator):
    _wildcard = re.compile(r'^https?://(?:[^./]+\.)+coherent\.global(?:/[^/?#]+)*(?:[?#].*)?$', re.IGNORECASE)

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(BaseUrlValidator, cls).__new__(cls)
        cls.instance.reset()
        return cls.instance

    def validate(self, value: Optional[str]):
        if is_str_empty(value):
            raise SparkError.sdk('base URL is required', value)

        if not self._wildcard.match(str(value)):
            raise SparkError.sdk('must be a Spark base URL <*.coherent.global>', value)

    def is_valid(self, value: Optional[str]) -> bool:
        try:
            self.validate(value)
            return True
        except SparkError as error:
            self._errors.append(error)
            return False


class Validators:
    empty_str = EmptyStringValidator
    positive_num = PositiveNumberValidator
    base_url = BaseUrlValidator



from __future__ import annotations

import math
import random
import re
import uuid
from typing import Any, List, Union, cast

from ._constants import DEFAULT_RETRY_INTERVAL, RETRY_RANDOMIZATION_FACTOR


def is_str(value: Any | None) -> bool:
    return isinstance(value, str)


def is_str_empty(value: Any | None) -> bool:
    return value is None or (isinstance(value, str) and value.strip() == '')


def is_str_not_empty(value: Any | None) -> bool:
    return isinstance(value, str) and value.strip() != ''


def is_int(value: Any | None) -> bool:
    if isinstance(value, int):
        return True
    try:
        int(str(value))
        return True
    except ValueError:
        return False


def is_positive_int(value: Any | None) -> bool:
    return is_int(value) and cast(int, value) > 0


def mask(value: str, start: int = 0, end: int = 4, char: str = '*') -> str:
    if not value or start < 0 or end < 0:
        return value
    return value[:start] + char * (len(value) - start - end) + value[-end:]


def sanitize_uri(url: str, leading: bool = False) -> str:
    sanitized = re.sub(r'/{2,}', '/', url)
    sanitized = sanitized.rstrip('/')
    return sanitized if leading else sanitized.lstrip('/')


def find_value_from_dict(key_path: str, data: dict, default: Any = None) -> Any:
    """
    Walks through a nested dictionary-based dataset and returns the value associated
    with the key path.
    """
    if not isinstance(data, dict):
        return default
    try:
        keys = key_path.split('.')
    except AttributeError:
        return default
    if not keys:
        return default

    # safe operations
    value = data.copy()
    for k in keys:
        value = cast(dict, value.get(k))
    return value


def get_retry_timeout(retries: int, interval=DEFAULT_RETRY_INTERVAL) -> float:
    randomization = random.uniform(0, 1) * RETRY_RANDOMIZATION_FACTOR
    return math.pow(2, retries) * interval * randomization


def join_list_str(value: Union[None, str, List[str]] = None, sep: str = ',') -> Union[str, None]:
    return sep.join(value) if isinstance(value, list) else value


def is_not_empty_list(value: Any | None) -> bool:
    return isinstance(value, list) and len(value) > 0


def get_uuid() -> str:
    return str(uuid.uuid4())



import logging
import re
import time
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Mapping, Optional, Union

from httpx import URL, Client, Headers, Request

from .._config import Config
from .._errors import SparkError
from .._utils import get_retry_timeout, get_uuid, is_str_empty, sanitize_uri
from .._version import about as sdk_info
from .._version import sdk_ua_header

__all__ = ['ApiResource', 'UriParams', 'Uri', 'HttpResponse']


class ApiResource:
    def __init__(self, config: Config):
        self.config = config
        self._client = Client()

        logging.basicConfig(
            format='[%(name)s] %(asctime)s [%(levelname)s] - %(message)s',
            datefmt='%m/%d/%Y, %H:%M:%S %p',
            level=logging.INFO,
        )
        # FIXME: Redefine HTTP handler to use the SDK config instead of httpx's.
        logger = 'cspark.sdk' if not self.config.logger else None  # hack to enable or disable logging
        self.logger = logging.getLogger(logger)

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def close(self):
        if not self._client.is_closed:
            self._client.close()

    @property
    def default_headers(self):
        return {
            **self.config.extra_headers,
            'User-Agent': sdk_info,
            'x-spark-ua': sdk_ua_header,
            'x-request-id': get_uuid(),
            'x-tenant-name': self.config.base_url.tenant,
        }

    def request(
        self,
        url: Union[str, 'Uri'],
        *,
        method: str = 'GET',
        headers: Mapping[str, str] = {},
        params: Optional[Mapping[str, str]] = None,
        body=None,
        form=None,
        files=None,
    ):
        url = url.value if isinstance(url, Uri) else url
        request = self._client.build_request(
            method,
            url,
            params=params,
            headers={**headers, **self.default_headers},
            data=form,
            json=body,
            files=files,
            timeout=self.config.timeout / 1000,
        )

        self.logger.debug(f'{method} {url}')
        return self.__fetch(request)

    def __fetch(self, request: Request, retries: int = 0) -> 'HttpResponse':
        request.headers.update(self.config.auth.as_header)
        response = self._client.send(request)
        status = response.status_code

        if status >= 400:
            if status == 401 and self.config.auth.type == 'oauth' and retries < self.config.max_retries:
                self.config.auth.oauth.retrieve_token(self.config)  # pyright: ignore[reportOptionalMemberAccess]
                return self.__fetch(request, retries + 1)

            if (status == 408 or status == 429) and retries < self.config.max_retries:
                self.logger.debug(f'retrying request due to status code {status}...')
                delay = get_retry_timeout(retries)
                time.sleep(delay)
                return self.__fetch(request, retries + 1)

            url = str(request.url)
            raise SparkError.api(
                status,
                {
                    'message': f'failed to fetch <{url}>',
                    'cause': {
                        'request': {
                            'url': url,
                            'method': request.method,
                            'headers': request.headers,
                            'body': request.content,
                        },
                        'response': {
                            'headers': response.headers,
                            'body': response.content,  # FIXME: cast to dict if possible
                            'raw': response.text,
                        },
                    },
                },
            )

        # otherwise, ok response
        http_response = {'status': status, 'data': None, 'buffer': response.content, 'headers': response.headers}

        content_type = response.headers.get('content-type', '')
        if 'application/json' in content_type:
            try:
                http_response['data'] = response.json()
            except Exception:
                http_response['data'] = response.text
        return HttpResponse(**http_response)


@dataclass
class HttpResponse:
    status: int
    data: Union[None, Any, str]
    buffer: bytes
    headers: Headers

    def copy_with(self, **kwargs) -> 'HttpResponse':
        return HttpResponse(
            status=kwargs.get('status', self.status),
            data=kwargs.get('data', self.data),
            buffer=kwargs.get('buffer', self.buffer),
            headers=kwargs.get('headers', self.headers),
        )


@dataclass(frozen=True)
class UriParams:
    folder: Optional[str] = None
    service: Optional[str] = None
    service_id: Optional[str] = None
    version: Optional[str] = None
    version_id: Optional[str] = None
    proxy: Optional[str] = None
    public: Optional[bool] = False

    @property
    def service_uri(self) -> str:
        """
        Returns the service URI locator's short format if available.
        folder/service[version?]
        """
        return Uri.encode(self, long=False)

    def pick(self, *args: str) -> 'UriParams':
        return UriParams(**{k: v for k, v in self.__dict__.items() if k in args})

    def omit(self, *args: str) -> 'UriParams':
        return UriParams(**{k: v for k, v in self.__dict__.items() if k not in args})

    def encode(self, long: bool = True) -> str:
        return Uri.encode(self, long=long)


class Uri:
    def __init__(self, url: URL):
        self._url = url

    @property
    def value(self) -> str:
        return str(self._url)

    @staticmethod
    def of(
        uri: Optional[UriParams] = None,
        *,
        base_url: str = '',
        version: str = 'api/v3',
        endpoint: str = '',
    ) -> 'Uri':
        uri = uri or UriParams()
        path = version
        if uri.public:
            path += '/public'

        if uri.version_id:
            path += f'/version/{uri.version_id}'
        elif uri.service_id:
            path += f'/service/{uri.service_id}'
        elif uri.folder and uri.service:
            path += f'/folders/{uri.folder}/services/{uri.service}'
        elif uri.proxy:
            path += f'/proxy/{sanitize_uri(uri.proxy)}'

        if endpoint and not uri.proxy:
            path += f'/{endpoint}'

        try:
            return Uri(URL(f'{base_url}/{path}'))
        except Exception as cause:
            raise SparkError.sdk('invalid URI params', uri) from cause

    @staticmethod
    def partial(
        resource: str,
        *,
        base_url: str = '',
        version: str = 'api/v3',
        endpoint: str = '',
    ) -> 'Uri':
        try:
            resource = sanitize_uri(resource)
            if version:
                resource = f'{version}/{resource}'
            if endpoint:
                resource += f'/{endpoint}'
            resource = f'{base_url}/{resource}'

            return Uri(URL(resource))
        except SparkError as error:
            raise SparkError.sdk(f'invalid service URI <{resource}>', error) from error
        except Exception as cause:
            raise SparkError(f'failed to build Spark endpoint from <{resource}>', cause) from cause

    @staticmethod
    def to_params(uri: Union[str, UriParams]) -> UriParams:
        return Uri.decode(uri) if isinstance(uri, str) else uri

    @staticmethod
    def decode(uri: str) -> UriParams:
        uri = re.sub('folders/', '', sanitize_uri(uri))
        uri = re.sub('services/', '', uri)
        match = re.match(r'^([^\/]+)\/([^[]+)(?:\[(.*?)\])?$', uri)
        if match:
            folder, service, version = match.groups()
            if folder == 'version':
                return UriParams(version_id=service)
            if folder == 'service':
                return UriParams(service_id=service)
            if folder == 'proxy':
                return UriParams(proxy=service)
            version = None if is_str_empty(version) else version
            return UriParams(folder=folder, service=service, version=version)
        return UriParams()

    @staticmethod
    def encode(uri: UriParams, long: bool = True) -> str:
        folder, service, version = uri.folder, uri.service, uri.version
        if uri.proxy:
            return f'proxy/{uri.proxy}'
        if uri.version_id:
            return f'version/{uri.version_id}'
        if uri.service_id:
            return f'service/{uri.service_id}'
        if folder and service:
            if long:
                return f'folders/{folder}/services/{service}{f"[{version}]" if version else ""}'
            return f'{folder}/{service}{f"[{version}]" if version else ""}'
        return ''

    @staticmethod
    def validate(uri: Union[str, UriParams], message: Optional[str] = None) -> UriParams:
        uri_params = Uri.to_params(uri)
        if is_str_empty(uri_params.service_uri):
            folder, service = uri_params.folder, uri_params.service
            if folder and not service:
                msg = message or 'service name is missing'
            elif not folder and service:
                msg = message or 'folder name is missing'
            else:
                msg = message or 'service uri locator is required'
            msg += ' :: a uri needs to be of these formats: \n\t- "folder/service[?version]" \n\t-'
            msg += ' "service/service_id" \n\t- "version/version_id" \n\t- "proxy/custom-endpoint"\n'
            raise SparkError.sdk(msg, uri)
        return uri_params

    def __str__(self) -> str:
        return self.value



from dataclasses import dataclass
from typing import Dict, cast

from .._config import Config
from ._base import ApiResource, Uri

__all__ = ['OAuth2', 'AccessToken']


class OAuth2(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)

    def get_access_token(self) -> 'AccessToken':
        url = Uri.of(base_url=self.config.base_url.oauth2, version='protocol', endpoint='openid-connect/token')
        body = {**self.config.auth.oauth.to_dict()}  # pyright: ignore[reportOptionalMemberAccess]
        body['grant_type'] = self.config.auth.oauth.flow  # pyright: ignore[reportOptionalMemberAccess]

        response = self.request(url, method='POST', form=body)
        data = cast(Dict, response.data)
        return AccessToken(
            access_token=str(data.get('access_token', '')),
            expires_in=int(data.get('expires_in', 0)),
            refresh_expires_in=int(data.get('refresh_expires_in', 0)),
            token_type=str(data.get('token_type', '')),
            not_before_policy=int(data.get('not-before-policy', 0)),
            scope=str(data.get('scope', '')),
        )


@dataclass(frozen=True)
class AccessToken:
    access_token: str
    expires_in: int
    refresh_expires_in: int
    token_type: str
    not_before_policy: int
    scope: str



import json
from typing import Any, Dict, List, Optional, Union

from .._config import Config
from .._constants import SPARK_SDK
from .._errors import SparkError
from .._utils import is_str_not_empty, join_list_str
from ._base import ApiResource, HttpResponse, Uri, UriParams

__all__ = ['Services', 'ServiceExecuted']


class Services(ApiResource):
    def __init__(self, config: Config):
        super().__init__(config)

    def execute(
        self,
        uri: Union[str, UriParams],
        *,
        response_format: str = 'alike',  # 'alike', 'typed', 'raw'
        # data for calculations
        inputs: Union[None, str, Dict[str, Any], List[Any]] = None,  # TODO: support `pandas.DataFrame`
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = SPARK_SDK,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3 (legacy)
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
    ):
        uri = Uri.validate(uri)

        executable = _ExecuteInputs(inputs)
        metadata = _ExecuteMeta(
            uri,
            is_batch=executable.is_batch,
            active_since=active_since,
            source_system=source_system,
            correlation_id=correlation_id,
            call_purpose=call_purpose,
            compiler_type=compiler_type,
            subservices=subservices,
            debug_solve=debug_solve,
            downloadable=downloadable,
            echo_inputs=echo_inputs,
            tables_as_array=tables_as_array,
            selected_outputs=selected_outputs,
            outputs_filter=outputs_filter,
        )

        if executable.is_batch:
            url = Uri.of(uri.pick('public'), base_url=self.config.base_url.full, version='api/v4', endpoint='execute')
            body = {'inputs': executable.inputs, **metadata.value}
        else:
            endpoint = '' if uri.version_id or uri.service_id else 'execute'
            url = Uri.of(uri, base_url=self.config.base_url.full, endpoint=endpoint)
            body = {'request_data': {'inputs': executable.inputs}, 'request_meta': metadata.value}

        response = self.request(url, method='POST', body=body)
        return ServiceExecuted(response, executable.is_batch, response_format)

    def get_schema(
        self, uri: Union[None, str, UriParams] = None, *, folder: Optional[str] = None, service: Optional[str] = None
    ):
        uri = UriParams(folder=folder, service=service) if uri is None else Uri.to_params(uri)
        endpoint = f'product/{uri.folder}/engines/get/{uri.service}'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)

        return self.request(url)

    def get_metadata(
        self,
        uri: Union[None, str, UriParams] = None,
        *,
        response_format: str = 'alike',  # 'alike', 'typed', 'raw'
        folder: Optional[str] = None,
        service: Optional[str] = None,
        service_id: Optional[str] = None,
        version_id: Optional[str] = None,
        proxy: Optional[str] = None,
        public: Optional[bool] = False,
    ):
        uri = (
            UriParams(folder, service, service_id, version_id=version_id, proxy=proxy, public=public)
            if uri is None
            else Uri.to_params(uri)
        )
        url = Uri.of(uri, base_url=self.config.base_url.full, endpoint='metadata')

        response = self.request(url)
        return ServiceExecuted(response, False, response_format)

    def get_versions(
        self, uri: Union[None, str, UriParams] = None, *, folder: Optional[str] = None, service: Optional[str] = None
    ):
        uri = UriParams(folder, service) if uri is None else Uri.to_params(uri)
        endpoint = f'product/{uri.folder}/engines/getversions/{uri.service}'
        url = Uri.of(base_url=self.config.base_url.value, version='api/v1', endpoint=endpoint)

        response = self.request(url)
        return response.copy_with(data=response.data.get('data', []) if isinstance(response.data, dict) else [])


class ServiceExecuted(HttpResponse):
    def __init__(self, response: HttpResponse, is_batch: bool, format: str = 'alike'):
        if format == 'raw':
            data = json.dumps(response.data)
        elif format == 'typed' or is_batch:
            data = response.data
        else:
            resp_data = response.data.get('response_data', {}) if isinstance(response.data, dict) else {}
            resp_meta = response.data.get('response_meta', {}) if isinstance(response.data, dict) else {}
            data = {
                'outputs': [resp_data.get('outputs')],
                'errors': [resp_data.get('errors')],
                'warnings': [resp_data.get('warnings')],
                **resp_meta,
            }
        super().__init__(response.status, data, response.buffer, response.headers)


class _ExecuteInputs:
    def __init__(self, data: Union[None, str, Dict[str, Any], List[Any]] = None):
        if data is None or (isinstance(data, list) and len(data) == 0):
            data = {}
        if is_str_not_empty(data):
            data = json.loads(str(data))

        self.inputs = data
        if isinstance(data, dict):
            self.length = 1
            self.is_batch = False
        elif isinstance(data, list):
            self.length = len(data)
            self.is_batch = True
        else:
            message = 'invalid data format\nexpected input data formats are string, dict or a list'
            raise SparkError.sdk(message, data)


class _ExecuteMeta:
    def __init__(
        self,
        uri: UriParams,
        is_batch: bool,
        *,
        # Metadata for calculations
        active_since: Optional[str] = None,
        source_system: Optional[str] = None,
        correlation_id: Optional[str] = None,
        call_purpose: Optional[str] = None,
        compiler_type: Optional[str] = None,
        subservices: Union[None, str, List[str]] = None,
        # Available only in v3 (legacy)
        debug_solve: Optional[bool] = None,
        downloadable: Optional[bool] = False,
        echo_inputs: Optional[bool] = False,
        tables_as_array: Union[None, str, List[str]] = None,
        selected_outputs: Union[None, str, List[str]] = None,
        outputs_filter: Optional[str] = None,
    ):
        self._uri = uri
        self._is_batch = is_batch
        self._active_since = active_since
        self._source_system = source_system or SPARK_SDK
        self._correlation_id = correlation_id

        self._call_purpose = (
            call_purpose
            if is_str_not_empty(call_purpose)
            else 'Sync Batch Execution'
            if is_batch
            else 'Single Execution'
        )
        self._compiler_type = (
            str(compiler_type).capitalize()
            if is_str_not_empty(compiler_type)
            and str(compiler_type).lower() in ('neuron', 'type3', 'type2', 'type1', 'xconnector')
            else 'Neuron'
        )

        self._subservices = join_list_str(subservices)
        self._debug_solve = debug_solve
        self._downloadable = downloadable
        self._echo_inputs = echo_inputs
        self._tables_as_array = join_list_str(tables_as_array)
        self._selected_outputs = join_list_str(selected_outputs)
        self._outputs_filter = outputs_filter

    @property
    def value(self) -> Dict[str, Any]:
        if self._is_batch:
            service_uri = self._uri.pick('folder', 'service', 'version').encode(long=False)
            return {
                'service': self._uri.service_id or service_uri or None,
                'version_id': self._uri.version_id,
                'version_by_timestamp': self._active_since,
                'subservice': self._subservices,
                'output': self._selected_outputs,
                'call_purpose': self._call_purpose,
                'source_system': self._source_system,
                'correlation_id': self._correlation_id,
            }

        return {
            # URI locator via metadata (v3 also supports URI in url path)
            'service_id': self._uri.service_id,
            'version_id': self._uri.version_id,
            'version': self._uri.version,
            # v3 expects extra metadata
            'transaction_date': self._active_since,
            'call_purpose': self._call_purpose,
            'source_system': self._source_system,
            'correlation_id': self._correlation_id,
            'array_outputs': self._tables_as_array,
            'compiler_type': self._compiler_type,
            'debug_solve': self._debug_solve,
            'excel_file': self._downloadable,
            'requested_output': self._selected_outputs,
            'requested_output_regex': self._outputs_filter,
            'response_data_inputs': self._echo_inputs,
            'service_category': self._subservices,
        }



from typing import Mapping, Optional

import cspark.sdk.resources as API

from ._config import BaseUrl, Config

__all__ = ['Client']


class Client:
    def __init__(
        self,
        *,
        base_url: Optional[str | BaseUrl] = None,
        oauth: Optional[Mapping[str, str] | str] = None,
        api_key: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[float] = None,
        max_retries: Optional[int] = None,
        retry_interval: Optional[float] = None,
        tenant: Optional[str] = None,
        env: Optional[str] = None,
        logger: Optional[bool] = True,
    ) -> None:
        self.config = Config(
            base_url=base_url,
            oauth=oauth,
            api_key=api_key,
            token=token,
            timeout=timeout,
            max_retries=max_retries,
            retry_interval=retry_interval,
            tenant=tenant,
            env=env,
            logger=logger,
        )

    @property
    def services(self) -> API.Services:
        return API.Services(self.config)

    @property
    def batches(self) -> API.Batches:
        return API.Batches(self.config)


class AsyncClient:
    pass
